import os
import json
import sys
import time
import tqdm
from redis import Redis
from chans import config
from chans import chans
from chans import util
from chans.chans.base import Chan


class Scraper:

    def __init__(
        self,
        chans: tuple[Chan, ...] = (chans.Ch2, chans.Ch4),
    ):
        self._chans: list[Chan] = chans
        self.redis = Redis.from_url(os.environ['REDIS_URL'], decode_responses=True)

    def chan_boards_key(self, chan: Chan) -> str:
        return f'{config.PREFIX}:boards:{chan.name}'

    def clean(self):
        util.delete_keys(f'{config.PREFIX}:', self.redis)

    def boards(self):
        pipeline = self.redis.pipeline()
        for chan in self._chans:
            key = self.chan_boards_key(chan)
            pipeline.delete(key)
            pipeline.sadd(key, *chan.boards())
        pipeline.execute()

    def posts(self):
        t = tqdm.tqdm(self._chans)
        for chan in self._chans:
            boards = self.redis.smembers(self.chan_boards_key(chan))
            for board in tqdm.tqdm(boards, leave=False):
                for post_raw in tqdm.tqdm(chan.board_posts_raw(board), leave=False):
                    post = chan.preprocess_post(post_raw)
                    if post.n_replies < config.MIN_REPLIES:
                        t.set_description(f'{chan.name} {board} {post.id} {post.comment[:100]} TOO FEW REPLIES')
                        continue
                    lifetime = time.time() - post.timestamp
                    if lifetime > config.EXPIRE_SECONDS:
                        t.set_description(f'{chan.name} {board} {post.id} {post.comment[:100]} EXPIRED')
                        continue
                    t.set_description(f'{chan.name} {board} {post.id} {post.comment[:100]} SAVE')

                    thread_posts_raw = chan.thread_posts_raw(board, post.id)

                    ttl = int(config.EXPIRE_SECONDS - lifetime)
                    key_raw = f'{config.PREFIX}:posts_raw:{chan.name}:{board}:{post.id}'
                    key_thread_posts_raw = f'{config.PREFIX}:thread_posts_raw:{chan.name}:{board}:{post.id}'
                    key = f'{config.PREFIX}:posts:{chan.name}:{board}:{post.id}'
                    pipeline = self.redis.pipeline()
                    pipeline.set(key_raw, json.dumps(post_raw, ensure_ascii=False))
                    pipeline.set(key_thread_posts_raw, json.dumps(thread_posts_raw, ensure_ascii=False))
                    pipeline.hset(key, mapping=post.dict())
                    pipeline.expire(key_raw, ttl)
                    pipeline.expire(key_thread_posts_raw, ttl)
                    pipeline.expire(key, ttl)
                    pipeline.execute()


    def __call__(self):
        pass


def parse():
    scraper = Scraper()
    scraper.boards()
    scraper.posts()

if __name__ == '__main__':
    # scraper = Scraper(chans=(chans.Ch4,))
    if len(sys.argv) == 2 and sys.argv[1] == 'clean':
        Scraper().clean()
        raise SystemExit
    parse()



import json
import abc
from typing import Any
from chans.schemas import Post
from chans import util
import collections
from redis import Redis
from chans import config


class Chan(abc.ABC):
    
    post_id_field: str
    comment_field: str

    @classmethod
    @abc.abstractmethod
    def boards(cls) -> list[str]:
        ...

    @classmethod
    @abc.abstractmethod
    def board_posts_raw(cls, board: str) -> list[dict]:
        ...

    @classmethod
    @abc.abstractmethod
    def preprocess_post(cls, post: dict) -> Post:
        ...

    @classmethod
    @abc.abstractmethod
    def thread_posts_raw(cls, board: str, thread_id: int) -> list[dict]:
        ...

    @classmethod
    @abc.abstractmethod
    def extract_replies(reply_raw: dict) -> set[int]:
        ...

    @classmethod
    def viral_replies(
        cls, 
        post: Post, 
        top_k: int = 3, 
        min_replies: int = 5,
        redis: Redis | None = None,
    ) -> list[Post]:
        """
        if redis is not None, it will be used to load thread data instead of making a request to API
        """
        if redis is not None:
            T = redis.get(f'{config.PREFIX}:thread_posts_raw:{cls.name}:{post.board}:{post.id}')
            if T is None:
                return []
            T = json.loads(T)
        T = cls.thread_posts_raw(post.board, post.id)
        # if not T:
            # return []
        assert T is not None, T
        T = T[1:] # delete OP post
        quotes = collections.Counter()
        for t in T:
            q = cls.extract_replies(t)
            quotes.update(q)
        most_quoted = {k: v for k, v in quotes.most_common(top_k + 1) if v >= min_replies}
    
        for t in T:
            if count := most_quoted.get(t[cls.post_id_field]):
                yield util.html2text(t[cls.comment_field], newline=False), count

    @classmethod
    def extract_comment(cls, post: dict[str, Any], candidate_fields: list[str]) -> str:
        candidate_fields = set(candidate_fields) & post.keys()
        candidates = {util.html2text(post[field]) for field in candidate_fields if post[field]}
        # todo: use difflib to drop similar comments
        return ' '.join(candidates)
    


  # if comment := thread.get('comment'):
        #     comment = util.html2text(comment, newline=True)
        #     nospace_title = [c for c in thread['title'] if not c.isspace()]
        #     nospace_comment = [c for c in comment if not c.isspace()]
        #     min_len = min(len(nospace_title), len(nospace_comment))
        #     ratio = difflib.SequenceMatcher(lambda x: x.isspace(), nospace_title[:min_len], nospace_comment[:min_len]).ratio()

        #     if ratio < 0.7:
        #         thread['title'] = thread['title']
        #     else:
        #         thread['title'] = comment[:500]



import operator
from redis import Redis
from chans.schemas import Post
from chans import config
import pipe21 as P


async def posts(
    redis: Redis,
    chans: frozenset[str] | None = None,
    boards: frozenset[str] | None = None,
    min_replies: int = 0,
    sortby: str = 'n_replies', 
    reverse: bool = True,
) -> list[Post]:
    keys = await redis.keys(f'{config.PREFIX}:posts:*')
    keys_filtered = []
    for key in keys:
        _, _, chan, board, _ = key.split(':')
        if chans is not None and chan not in chans:
            continue
        if boards is not None and board not in boards:
            continue
        keys_filtered.append(key)
    pipeline = redis.pipeline()
    for key in keys_filtered:
        pipeline.hgetall(key)
    posts = await pipeline.execute()
    posts = [Post(**post) for post in posts]
    return (
        posts
        | P.Filter(lambda post: post.n_replies >= min_replies)
        | P.Sorted(key = operator.attrgetter(sortby), reverse=reverse)
        | P.Pipe(list)
    )



import time
import random
import difflib
import shutil
import pipe21 as P
from chans import util
from chans.util import color_string
from chans.util import color_i
from chans import chans
from chans import util



def main():
    while True:
        threads = (
            util.get_threads()
            | P.Filter(lambda thread: thread['lifetime_seconds'] < 60 * 60 * 1.5)
            | P.Pipe(list)
            | P.Apply(random.shuffle)
        )

        for i, thread in enumerate(threads, start=1):
            title = thread['title']
            # title = util.html2text(title, newline=True)
            posts_count = thread['posts_count']
            sleep_time = round(-12 * 0.7 ** (0.03 * posts_count) + 12, 1)

            title_color = util.color(posts_count)
            derivative = int(posts_count/thread['lifetime_seconds'] * 1000)


            print('{:<14} {:<10} {} {:>10} {:>18} {:>10} {}'.format(
                color_string.YELLOW(f'{i}..{len(threads)}'),
                color_string.BLUE(f"{thread['board']:<4}"),
                posts_count and f'⬆ {derivative:<4}' or '',
                color_string.WHITE(f"{posts_count:>4} comments"),
                util.format_time(thread['timestamp']),
                color_string.CYAN(f'{sleep_time} sleep'),
                thread['url'],
            ))


            if comment := thread.get('comment'):
                comment = util.html2text(comment, newline=True)

                nospace_title   = [c for c in title   if not c.isspace()]
                nospace_comment = [c for c in comment if not c.isspace()]
                min_len = min(len(nospace_title), len(nospace_comment))

                ratio = difflib.SequenceMatcher(lambda x: x.isspace(), nospace_title[:min_len], nospace_comment[:min_len]).ratio()

                if ratio < 0.7:
                    print(color_i(title, title_color))
                print(color_i(comment[:500], title_color))
            else:
                print(color_i(title, title_color))

            t0 = time.time()
            for viral_reply, count in {'4ch': chans.Ch4, '2ch': chans.Ch2}[thread['chan']].viral_replies(thread, 3, min_replies=5):
                print(color_string.WHITE(f'[{count} replies]'), end=' ')
                print(color_i(viral_reply, title_color))

            print(color_i('─'*shutil.get_terminal_size().columns, 236))
            time.sleep(max(sleep_time - (time.time() - t0), 0))

if __name__ == '__main__':
    main()



import collections
import re
from chans import util
from chans.chans.base import Chan
from chans.schemas import Post
from typing import Any

import pipe21 as P

class Ch4(Chan):
    name = '4ch'
    host = 'https://4chan.org'
    post_id_field = 'no'
    comment_field = 'com'

    @classmethod
    def boards(cls) -> list[str]:
        return [board['board'] for board in util.retry_request('https://a.4cdn.org/boards.json').json()['boards']]

    @classmethod
    def board_posts_raw(cls, board: str) -> list[dict]:
        return (
            util.retry_request(f'https://a.4cdn.org/{board}/catalog.json').json()
            | P.FlatMap(lambda x: x['threads'])
            | P.ApplyMap(util.setitem('board', board))
            | P.Pipe(list)
        )
    
    @classmethod
    def preprocess_post(cls, post: dict[str, Any]) -> Post:
        id_ = post[cls.post_id_field]
        post['id'] = id_
        post['timestamp'] = post['time']
        post['url'] = f"https://boards.4chan.org/{post['board']}/thread/{id_}"
        post['comment'] = cls.extract_comment(post, candidate_fields=['sub', 'com', 'comment'])
        post['chan'] = cls.name
        post['n_replies'] = post['replies']
        return Post(**post)
    
    @classmethod
    def thread_posts_raw(cls, board: str, thread_id: int) -> list[dict]:
        return util.retry_request(f"https://a.4cdn.org/{board}/thread/{thread_id}.json").json()['posts']

    def extract_replies(post_raw: dict) -> set[int]:
        if com := post_raw.get('com'):
            return {int(x) for x in re.findall(r'(?<=href="#p)\d+(?=" class="quotelink")', com)}
        return set()

    # @classmethod
    # def viral_replies(cls, post: Post, top_k: int = 3, min_replies: int = 5) -> list[Post]:
    #     T = cls.thread_posts_raw(post.board, post.id)
    #     # if not T:
    #         # return []
    #     T = T[1:] # delete OP post

    #     quotes = collections.Counter()

    #     for t in T:
    #         if com := t.get('com'):
    #             q = map(int, set(re.findall(r'(?<=href="#p)\d+(?=" class="quotelink")', com)))
    #             quotes.update(q)

    #     t0 = T[0]['no']
    #     most_quoted = {k: v for k, v in quotes.most_common(top_k + 1) if v >= min_replies}
    #     if t0 in most_quoted:
    #         del most_quoted[t0]

    #     for t in T:
    #         if (count := most_quoted.get(t['no'])) and (com := t.get('com')):
    #             yield util.html2text(com, newline=False), count



from fastapi import FastAPI
from fastapi import Request
from fastapi import Depends
from fastapi import WebSocket
from fastapi import HTTPException
from fastapi.responses import HTMLResponse
from fastapi.responses import FileResponse
from fastapi.responses import RedirectResponse
from fastapi.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates
from pathlib import Path
import subprocess
import asyncio
from redis.asyncio import Redis
import threading


from chans import chans
from chans import util
from chans import data
from chans import config
from chans import parser
from chans.stream import prepare_posts
from chans.dependencies import get_redis_async


parser_thread = threading.Thread(target=parser.parse)
parser_thread.start()


static_folder = Path(__file__).parent / 'static'
templates = Jinja2Templates(directory=static_folder / 'templates')
templates.env.filters['format_time'] = util.format_time

app = FastAPI()
app.mount('/static/', StaticFiles(directory=static_folder), name='static')


@app.get('/favicon.ico', response_class=HTMLResponse)
async def favicon():
    return FileResponse(static_folder / 'favicon.ico')


@app.get('/', response_class=HTMLResponse)
async def get_chan(
    request: Request,
    min_replies: int = 10,
    redis: Redis = Depends(get_redis_async),
):
    posts = await data.posts(redis, min_replies=min_replies)
    return templates.TemplateResponse(
        'posts.j2', {
            'request': request,
            'posts': posts,
        },
    )


@app.get('/stream', response_class=HTMLResponse)
def stream(request: Request):
    return templates.TemplateResponse(
        'stream.j2', {
            'request': request,
            'port': config.PORT,
        },
    )

@app.websocket("/ws")
async def websocket_endpoint(
    websocket: WebSocket,
    redis: Redis = Depends(get_redis_async),
):
    posts = await data.posts(redis)
    posts = prepare_posts(posts)
    await websocket.accept()
    for post in posts:
        card_html = templates.get_template('card.j2').render(post=post)
        await websocket.send_text(card_html)
        await asyncio.sleep(post.sleep_time)


@app.get('/settings', response_class=HTMLResponse)
async def settings(
    request: Request,
):
    return templates.TemplateResponse(
        'settings.j2', {
            'request': request,
        },
    )


def main():
    cmd = ["uvicorn", "chans.server:app", "--host", "0.0.0.0", "--port", config.PORT]
    subprocess.run(cmd)

if __name__ == '__main__':
    main()



import time
import random
import difflib
import pipe21 as P
from chans import util
from chans.util import color_string
from chans.util import color_i
from chans import chans
from chans.chans.base import Chan
from chans import util
from chans import schemas


def sleep_time(posts_count) -> float:
    return -12 * 0.7 ** (0.03 * posts_count) + 12


def prepare_posts(posts: list[schemas.Post]):
    random.shuffle(posts)

    for post in posts:
        t0 = time.time()
        # thread['title_color'] = util.color(thread['posts_count'])
        lifetime_seconds = time.time() - post.timestamp
        derivative = int(post.n_replies / lifetime_seconds * 1000)

        print(post)
        viral_replies = []
        cls: Chan = {'4ch': chans.Ch4, '2ch': chans.Ch2}[post.chan]
        for comment, n_replies in cls.viral_replies(post, top_k=3, min_replies=5):
            viral_replies.append(schemas.ViralReply(comment=comment, n_replies=n_replies))

        yield schemas.StreamPost(
            **post.dict(),
            derivative=derivative,
            sleep_time=max(sleep_time(post.n_replies) - (time.time() - t0), 0),
            viral_replies=viral_replies,
        )



from collections.abc import Callable
from pydantic import BaseModel


class Post(BaseModel):
    id: int
    timestamp: int
    comment: str
    chan: str
    board: str
    n_replies: int
    url: str


# class ViralReply(Post):
class ViralReply(BaseModel):
    comment: str
    n_replies: int


class StreamPost(Post):
    derivative: int
    sleep_time: float
    viral_replies: list[ViralReply]



import contextlib
import os

from redis import Redis
from redis.asyncio import Redis as RedisAsync

def get_redis():
    r = Redis.from_url(os.environ['REDIS_URL'], decode_responses=True)
    yield r
    r.close()

def get_redis_async():
    r = RedisAsync.from_url(os.environ['REDIS_URL'], decode_responses=True)
    yield r
    r.close()

get_redis_cm = contextlib.contextmanager(get_redis)
get_redis_async_cm = contextlib.contextmanager(get_redis_async)



import collections
import re
from chans import util
from chans.schemas import Post
from chans.schemas import ViralReply
from typing import Any

from chans.chans.base import Chan
from chans import util


class Ch2(Chan):
    name = '2ch'
    host = 'https://2ch.hk'
    post_id_field = 'num'
    comment_field = 'comment'

    @classmethod
    def boards(cls) -> list[str]:
        return [board['id'] for board in util.retry_request(f'{cls.host}/index.json').json()['boards']]

    @classmethod
    def board_posts_raw(cls, board: str) -> list[dict]:
        return util.retry_request(f'{cls.host}/{board}/catalog.json').json()['threads']
    
    @classmethod
    def preprocess_post(cls, post: dict[str, Any]) -> Post:
        id_ = post[cls.post_id_field]
        post['id'] = id_
        post['url'] = f"{cls.host}/{post['board']}/res/{id_}.html"
        post['comment'] = cls.extract_comment(post, candidate_fields=['subject', 'comment'])
        post['chan'] = cls.name
        post['n_replies'] = post['posts_count']
        return Post(**post)

    @classmethod
    def thread_posts_raw(cls, board: str, thread_id: int) -> list[dict]:
        return util.retry_request(f"{cls.host}/{board}/res/{thread_id}.json").json()['threads'][0]['posts']
    
    def extract_replies(post_raw: dict) -> set[int]:
        return {int(x) for x in re.findall(r'(?<=#)\d+(?=" class="post-reply-link")', post_raw['comment'])}

    # @classmethod
    # def viral_replies(cls, post: Post, top_k: int = 3, min_replies: int = 5) -> list[Post]:
    #     T = cls.thread_posts_raw(post.board, post.id)
    #     # if not T:
    #         # return []
    #     T = T[1:] # delete OP post
        
    #     quotes = collections.Counter()

    #     for t in T:
    #         com = t['comment']
    #         q = map(int, set(re.findall(r'(?<=#)\d+(?=" class="post-reply-link")', com)))
    #         quotes.update(q)

    #     t0 = T[0]['num']
    #     most_quoted = {k: v for k, v in quotes.most_common(top_k + 1) if v >= min_replies}
    #     if t0 in most_quoted: # delete OP post
    #         del most_quoted[t0]

    #     for t in T:
    #         if count := most_quoted.get(t['num']):
    #             yield util.html2text(t['comment'], newline=False), count



import datetime
import itertools
import re
import html
import time
import humanize
from redis import Redis
from typing import Any
import requests
from tenacity import retry
from tenacity import stop_after_attempt
from tenacity import wait_fixed
from tenacity import TryAgain

from chans.dependencies import get_redis_cm
from chans import chans
from chans import config


def delete_keys(prefix: str, redis):
    pipeline = redis.pipeline()
    for key in redis.keys(prefix + '*'):
        pipeline.delete(key)
    _ = pipeline.execute()


def setitem(key, value):
    def inner(x):
        x[key] = value
    return inner



def log_error_to_redis(retry_state):
    error_data, = retry_state.outcome._exception.args
    timestamp = round(time.time() * 1000)
    print(timestamp, error_data)
    with get_redis_cm() as redis:
        redis.hset(f'{config.ERRORS}:{timestamp}', mapping=error_data)


@retry(
    wait=wait_fixed(2),
    stop=stop_after_attempt(3),
    retry_error_callback=log_error_to_redis, 
)
def retry_request(url: str) -> dict:
    r = requests.get(url)
    if not r.ok:
        raise TryAgain({
            'url': url,
            'status_code': r.status_code,
            'text': r.text,
        })
    return r


def html2text(htm, newline=False):
    ret = html.unescape(htm)
    ret = ret.translate({8209: ord('-'), 8220: ord('"'), 8221: ord('"'), 160: ord(' '),})
    ret = re.sub(r"\s", " ", ret, flags = re.MULTILINE)
    if newline:
        ret = re.sub("<br>|<br />|</p>|</div>|</h\d>", '\n', ret, flags = re.IGNORECASE)
    else:
        ret = re.sub("<br>|<br />|</p>|</div>|</h\d>", ' ', ret, flags = re.IGNORECASE)
    ret = re.sub('<.*?>', ' ', ret, flags=re.DOTALL)
    ret = re.sub(r"  +", " ", ret)
    return ret


def color(v):
    if   v <   5: c = 236
    elif v <  10: c = 237
    elif v <  15: c = 238
    elif v <  20: c = 239
    elif v <  25: c = 240
    elif v <  30: c = 241
    elif v <  35: c = 242
    elif v <  40: c = 243
    elif v <  45: c = 244
    elif v <  50: c = 245
    elif v <  55: c = 246
    elif v <  60: c = 247
    elif v <  65: c = 248
    elif v <  70: c = 249
    elif v <  75: c = 250
    elif v <  80: c = 251
    elif v <  85: c = 252
    elif v <  90: c = 253
    elif v <  95: c = 254
    elif v < 100: c = 255
    elif v < 200: c = 82  # green
    elif v < 300: c = 87  # blue
    elif v < 400: c = 226 # yellow
    elif v < 500: c = 208 # orange
    elif v>= 500: c = 196 # red
    return c

class color_string:
    BLACK     = lambda s: '\033[30m' + str(s) + '\033[0m'
    RED       = lambda s: '\033[31m' + str(s) + '\033[0m'
    GREEN     = lambda s: '\033[32m' + str(s) + '\033[0m'
    YELLOW    = lambda s: '\033[33m' + str(s) + '\033[0m'
    BLUE      = lambda s: '\033[34m' + str(s) + '\033[0m'
    MAGENTA   = lambda s: '\033[35m' + str(s) + '\033[0m'
    CYAN      = lambda s: '\033[36m' + str(s) + '\033[0m'
    WHITE     = lambda s: '\033[37m' + str(s) + '\033[0m'
    UNDERLINE = lambda s: '\033[4m'  + str(s) + '\033[0m'


def color_i(s, i):
    return f"\x1b[38;5;{i}m{s}\033[0m"


def to_local_time(dt: datetime.datetime, timezone: datetime.timezone):
    if dt.tzinfo is not None:
        raise ValueError('pass timezone as separate argument')
    dt = dt.replace(tzinfo=timezone)
    return datetime.datetime.fromtimestamp(dt.timestamp())


def format_time(
    t: datetime.datetime | int,
    absolute: bool = False,
    pad: bool = False,
    timezone: datetime.timezone | None = None,
) -> str:
    if isinstance(t, int):
        t = datetime.datetime.fromtimestamp(t)
    if timezone is not None:
        t = to_local_time(t, timezone)
    if absolute or (datetime.datetime.utcnow() - t).days > 30:  # noqa: PLR2004
        return t.strftime('%Y %b %d %H:%M')
    t = datetime.datetime.fromtimestamp(t.timestamp())
    out = humanize.naturaltime(t)
    if pad:
        out = out.rjust(17)
    return out
